create or replace
PACKAGE BODY sbip_schedule AS
/*****************************************************************************/
--<GLOBAL_PRIVATE_DATA_STRUCTURES>
   TYPE calendar_rt IS RECORD (
      frequency        VARCHAR2(10),
      interval         PLS_INTEGER,
      calendars_used   BOOLEAN,
      bysecond         DBMS_SCHEDULER.BYLIST,
      byminute         DBMS_SCHEDULER.BYLIST,
      byhour           DBMS_SCHEDULER.BYLIST,
      byday_days       DBMS_SCHEDULER.BYLIST,
      byday_occurrence DBMS_SCHEDULER.BYLIST,
      bymonthday       DBMS_SCHEDULER.BYLIST,
      byyearday        DBMS_SCHEDULER.BYLIST,
      byweekno         DBMS_SCHEDULER.BYLIST,
      bymonth          DBMS_SCHEDULER.BYLIST
   );
   TYPE schedule_rt IS RECORD(
      sched_start_date      VARCHAR2(12),
      sched_end_date        VARCHAR2(12),
      sched_calendar_string VARCHAR2(32767),
      sched_duration        PLS_INTEGER,
      start_date_dt         DATE,
      start_date_ts         TIMESTAMP,
      end_date_dt           DATE,
      end_date_ts           TIMESTAMP,
      calendar_rec          CALENDAR_RT
   );
--</GLOBAL_PRIVATE_DATA_STRUCTURES>
--
--<GLOBAL_PRIVATE_VARIABLES>
   g_schedule_rec SCHEDULE_RT;
--</GLOBAL_PRIVATE_VARIABLES>
/******************************************************************************/
PROCEDURE parse_start_date(
   schedule_string_in IN VARCHAR2
)

IS

BEGIN

   g_schedule_rec.sched_start_date := SUBSTR(schedule_string_in, 1, 12);
   g_schedule_rec.start_date_dt := TO_DATE(g_schedule_rec.sched_start_date, 'YYYYMMDDHH24MI');
   g_schedule_rec.start_date_ts := CAST(g_schedule_rec.start_date_dt AS TIMESTAMP);
   
END;
/******************************************************************************/
PROCEDURE parse_end_date(
   schedule_string_in IN VARCHAR2
)

IS

BEGIN

   g_schedule_rec.sched_end_date := SUBSTR(schedule_string_in, INSTR(schedule_string_in, '|') + 1, INSTR(schedule_string_in, '|', 1, 2) - (INSTR(schedule_string_in, '|') + 1));
   
   IF g_schedule_rec.sched_end_date IS NOT NULL
   THEN
      g_schedule_rec.end_date_dt := TO_DATE(SUBSTR(g_schedule_rec.sched_end_date, 1, 8)
         || TO_CHAR(g_schedule_rec.start_date_ts, 'HH24MI'), 'YYYYMMDDHH24MI');--Set the end date's time the same as the start dates
      g_schedule_rec.end_date_ts := CAST(g_schedule_rec.end_date_dt AS TIMESTAMP);
   ELSE
      g_schedule_rec.end_date_dt := NULL;
      g_schedule_rec.end_date_ts := NULL;
   END IF;
   
END;
/******************************************************************************/
PROCEDURE parse_duration(
   schedule_string_in IN VARCHAR2
)

IS

BEGIN

   g_schedule_rec.sched_duration := SUBSTR(schedule_string_in, INSTR(schedule_string_in, '|', 1, 2) + 1, INSTR(schedule_string_in, '|', 1, 3) - (INSTR(schedule_string_in, '|', 1, 2) + 1));

END;
/******************************************************************************/
PROCEDURE parse_calendar_string(
   schedule_string_in IN VARCHAR2
)

IS

BEGIN

   g_schedule_rec.sched_calendar_string := SUBSTR(schedule_string_in, INSTR(schedule_string_in, '|', 1, 3) + 1);
   
   --DBMS_SCHEDULER mappings seem off for byweekno, bymonthday, and byyearday - this has been adjusted to compensate
   dbms_scheduler.resolve_calendar_string(
      calendar_string  => g_schedule_rec.sched_calendar_string,
      frequency        => g_schedule_rec.calendar_rec.frequency,
      interval         => g_schedule_rec.calendar_rec.interval,
      bysecond         => g_schedule_rec.calendar_rec.bysecond,
      byminute         => g_schedule_rec.calendar_rec.byminute,
      byhour           => g_schedule_rec.calendar_rec.byhour,
      byday_days       => g_schedule_rec.calendar_rec.byday_days,
      byday_occurrence => g_schedule_rec.calendar_rec.byday_occurrence,
      bymonthday       => g_schedule_rec.calendar_rec.byweekno, --adjusted
      byyearday        => g_schedule_rec.calendar_rec.bymonthday, --adjusted
      byweekno         => g_schedule_rec.calendar_rec.byyearday, --adjusted
      bymonth          => g_schedule_rec.calendar_rec.bymonth
   );
   
END;
/******************************************************************************/
PROCEDURE parse_all (
   schedule_string_in IN VARCHAR2
)

IS

BEGIN

   g_schedule_rec := NULL;
   parse_start_date(schedule_string_in);
   parse_end_date(schedule_string_in);
   parse_duration(schedule_string_in);
   parse_calendar_string(schedule_string_in);
   
END;
/******************************************************************************/
FUNCTION dates (
   schedule_string_in  IN VARCHAR2,
   occurrence_limit_in IN PLS_INTEGER := 1000,
   window_start_in     IN DATE := NULL,
   window_end_in       IN DATE := NULL
)

   RETURN DATE_NTT
      
   PIPELINED
   
   DETERMINISTIC
   
IS

   l_dates             DATE_NTT := DATE_NTT();
   l_window_start      TIMESTAMP;
   l_window_end        TIMESTAMP;
   l_return_date_after TIMESTAMP;
   l_next_run_date     TIMESTAMP;
   l_limit             PLS_INTEGER;
   l_idx               PLS_INTEGER := 0;
   
BEGIN

   parse_all(schedule_string_in);
   
   l_limit := NVL(occurrence_limit_in, 1000);
   l_window_start := NVL(CAST(window_start_in AS TIMESTAMP), g_schedule_rec.start_date_ts);
   
   IF window_end_in IS NOT NULL
   THEN
      l_window_end := TO_TIMESTAMP(TO_CHAR(window_end_in, 'YYYYMMDD') || TO_CHAR(g_schedule_rec.start_date_ts, 'HH24MISS'), 'YYYYMMDDHH24MISS'); --Set the window end time the same as the start date's
   END IF;
   
   l_next_run_date := g_schedule_rec.start_date_ts;
   l_return_date_after := g_schedule_rec.start_date_ts - NUMTODSINTERVAL(1, 'SECOND'); --One second shaved from start to make sure the first occurance is included.
   
   IF g_schedule_rec.sched_calendar_string IS NULL
   THEN
      PIPE ROW (g_schedule_rec.start_date_dt);
   ELSIF g_schedule_rec.end_date_ts IS NULL AND window_start_in IS NULL AND window_end_in IS NULL
   THEN
      WHILE l_idx < l_limit
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         PIPE ROW (l_next_run_date);
         l_idx := l_idx + 1;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NOT NULL AND window_start_in IS NULL AND window_end_in IS NULL
   THEN
      WHILE l_idx < l_limit
         AND l_next_run_date <= g_schedule_rec.end_date_ts
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date <= g_schedule_rec.end_date_ts
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NULL AND window_start_in IS NOT NULL AND window_end_in IS NULL
   THEN
      WHILE l_idx < l_limit
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date >= l_window_start
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NULL AND window_start_in IS NULL AND window_end_in IS NOT NULL
   THEN
      WHILE l_idx < l_limit
         AND l_next_run_date <= l_window_end
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date <= l_window_end
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NOT NULL AND window_start_in IS NULL AND window_end_in IS NOT NULL
   THEN
      WHILE l_idx < l_limit
         AND l_next_run_date <= g_schedule_rec.end_date_ts
         AND l_next_run_date <= l_window_end
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date <= g_schedule_rec.end_date_ts
            AND l_next_run_date <= l_window_end
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NOT NULL AND window_start_in IS NOT NULL AND  window_end_in IS NULL
   THEN
      WHILE l_idx < l_limit
         AND l_next_run_date <= g_schedule_rec.end_date_ts
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date <= g_schedule_rec.end_date_ts
            AND l_next_run_date >= l_window_start
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSIF g_schedule_rec.end_date_ts IS NULL AND window_start_in IS NOT NULL AND  window_end_in IS NOT NULL
   THEN
      WHILE l_idx < l_limit
         AND l_next_run_date <= l_window_end
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date >= l_window_start
            AND l_next_run_date <= l_window_end
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   ELSE -- g_schedule_rec.end_date IS NOT NULL AND window_start_in IS NOT NULL AND  window_end_in IS NOT NULL
      WHILE l_idx < l_limit
         AND l_next_run_date <= g_schedule_rec.end_date_ts
         AND l_next_run_date <= l_window_end
      LOOP
         dbms_scheduler.evaluate_calendar_string(
            calendar_string   => g_schedule_rec.sched_calendar_string,
            start_date        => g_schedule_rec.start_date_ts,
            return_date_after => l_return_date_after,
            next_run_date     => l_next_run_date
         );
         
         IF l_next_run_date >= l_window_start
            AND l_next_run_date <= g_schedule_rec.end_date_ts
            AND l_next_run_date <= l_window_end
         THEN
            PIPE ROW (l_next_run_date);
            l_idx := l_idx + 1;
         END IF;
         
         l_return_date_after := l_next_run_date;
      END LOOP;
   END IF;
   
   RETURN;
   
END dates;
/******************************************************************************/
FUNCTION start_date (
   schedule_string_in IN VARCHAR2
)
   
   RETURN DATE
   
   DETERMINISTIC
   
IS

BEGIN
  
   parse_start_date(schedule_string_in);
   
   RETURN g_schedule_rec.start_date_dt;
   
END;
/******************************************************************************/
FUNCTION end_date (
   schedule_string_in IN VARCHAR2
)
   
   RETURN DATE
   
   DETERMINISTIC
   
IS

BEGIN
  
   parse_end_date(schedule_string_in);
   
   RETURN g_schedule_rec.end_date_dt;
   
END;
/******************************************************************************/
FUNCTION duration_seconds (
   schedule_string_in IN VARCHAR2
)
   
   RETURN PLS_INTEGER
   
   DETERMINISTIC
   
IS

BEGIN
  
   parse_duration(schedule_string_in);
   
   RETURN g_schedule_rec.sched_duration;
   
END;
/******************************************************************************/
FUNCTION calendar_summary (
   schedule_string_in IN VARCHAR2
)
   
   RETURN VARCHAR2
   
   DETERMINISTIC
   
IS

   l_retval VARCHAR2(32767);
   
   FUNCTION ordinal(
      p_number NUMBER
   )
   
      RETURN VARCHAR2
      
   IS
   
      l_retval      VARCHAR2(32767);
      l_last_number PLS_INTEGER;
      
   BEGIN
   
      l_last_number := SUBSTR(p_number, -1);
      
      IF l_last_number > 3
      THEN
         l_retval := p_number || 'th';
      ELSIF l_last_number = 1
      THEN
         l_retval := p_number || 'st';
      ELSIF l_last_number = 2
      THEN
         l_retval := p_number || 'nd';
      ELSIF l_last_number = 3
      THEN
         l_retval := p_number || 'rd';
      END IF;
      
      RETURN l_retval;
      
   END;
   
BEGIN
  
   IF schedule_string_in IS NOT NULL
   THEN
      parse_all(schedule_string_in);
      
      IF g_schedule_rec.calendar_rec.frequency = DBMS_SCHEDULER.DAILY
      THEN
         IF g_schedule_rec.calendar_rec.interval = 1
         THEN
            l_retval := 'Daily';
         ELSE
            l_retval := 'Every ' || g_schedule_rec.calendar_rec.interval || ' days';
         END IF;
      ELSIF g_schedule_rec.calendar_rec.frequency = DBMS_SCHEDULER.WEEKLY
      THEN
         IF g_schedule_rec.calendar_rec.interval = 1
         THEN
            l_retval := 'Weekly on ';
         ELSE
            l_retval := 'Every ' || g_schedule_rec.calendar_rec.interval || ' weeks on ';
         END IF;
         
         FOR x IN 1 .. g_schedule_rec.calendar_rec.byday_days.count
         LOOP
            IF g_schedule_rec.calendar_rec.byday_days.count > 1
               AND x = g_schedule_rec.calendar_rec.byday_days.count
            THEN
               l_retval := l_retval || 'and ';
            END IF;
            
            IF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.SUNDAY
            THEN
               l_retval := l_retval || 'Sunday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.MONDAY
            THEN
               l_retval := l_retval || 'Monday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.TUESDAY
            THEN
               l_retval := l_retval || 'Tuesday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.WEDNESDAY
            THEN
               l_retval := l_retval || 'Wednesday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.THURSDAY
            THEN
               l_retval := l_retval || 'Thursday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.FRIDAY
            THEN
               l_retval := l_retval || 'Friday, ';
            ELSIF g_schedule_rec.calendar_rec.byday_days(x) = DBMS_SCHEDULER.SATURDAY
            THEN
               l_retval := l_retval || 'Saturday, ';
            END IF;
         END LOOP;
            
         l_retval := RTRIM(l_retval, ', ');
         IF g_schedule_rec.calendar_rec.byday_days.count = 2
         THEN
            l_retval := REPLACE(l_retval, ',', '');
         END IF;
      ELSIF g_schedule_rec.calendar_rec.frequency = DBMS_SCHEDULER.MONTHLY
      THEN
         IF g_schedule_rec.calendar_rec.interval = 1
         THEN
            l_retval := 'Monthly on the ';
         ELSE
            l_retval := 'Every ' || g_schedule_rec.calendar_rec.interval || ' months on the ';
         END IF;
         
         IF g_schedule_rec.calendar_rec.bymonthday.exists(1)
         THEN
            l_retval := l_retval || ordinal(g_schedule_rec.calendar_rec.bymonthday(1));
         ELSIF g_schedule_rec.calendar_rec.byday_days.exists(1)
         THEN
            IF  g_schedule_rec.calendar_rec.byday_occurrence(1) = -1
            THEN
               l_retval := l_retval || 'last ';
            ELSE
               l_retval := l_retval || ordinal(g_schedule_rec.calendar_rec.byday_occurrence(1)) || ' ';
            END IF;
            
            l_retval := l_retval || TO_CHAR(g_schedule_rec.start_date_dt, 'Day');
         END IF;
      ELSIF g_schedule_rec.calendar_rec.frequency = DBMS_SCHEDULER.YEARLY
      THEN
         l_retval := 'Yearly on the '
            || ordinal(g_schedule_rec.calendar_rec.bymonthday(1))
            || ' of '
            || TO_CHAR(g_schedule_rec.start_date_dt, 'Month');
      END IF;
   END IF;
   
   RETURN l_retval;
   
END;
/******************************************************************************/
END sbip_schedule;